//
// Created by webpigeon on 23/10/2021.
//

#ifndef FGGL_ECS3_PROTOTYPE_WORLD_H
#define FGGL_ECS3_PROTOTYPE_WORLD_H

#include <map>
#include <fggl/ecs3/types.hpp>

/**
 * A component based implementation of a game world.
 *
 * This is not a true ECS but exposes a similar API to it for testing (with a lot less headaches).
 */
namespace fggl::ecs3::prototype {


    class Entity {
        public:
            Entity(entity_t id) : m_id() {};
            ~Entity() = default;

            template<typename C>
            C* add() {
                C* ptr = new C();
                m_components[Component<C>::typeID()] = ptr;
                return ptr;
            }

            template<typename C>
            C* set(const C* ptr) {
                C* newPtr = new C(*ptr);
                m_components[Component<C>::typeID()] = newPtr;
                return newPtr;
            }

            template<typename C>
            C* get() {
                void* ptr = m_components.at(Component<C>::typeID());
                return (C*)ptr;
            }

            bool hasComponents(std::vector<component_type_t>& Cs) {
                for (auto c : Cs) {
                    if ( m_components.find(c) == m_components.end() ) {
                        return false;
                    }
                }
                return true;
            }

        private:
            entity_t m_id;
            std::map<component_type_t, void*> m_components;
    };


    class World {
        public:
            World(TypeRegistry& reg) : m_types(reg), m_next(0), m_entities() {};
            ~World() = default;

            entity_t create() {
                auto nextID = m_next++;
                m_entities.emplace(nextID, nextID);
                return nextID;
            }

            void destroy(entity_t entity) {
                // TOOD resolve and clean components
                m_entities.erase(entity);
            }

            template<typename... Cs>
            std::vector<entity_t> findMatching() {
                // construct the key
                std::vector<ecs::component_type_t> key;
                (key.push_back( Component<Cs>::typeID() ), ...);

                // entities
                std::vector<entity_t> entities{};
                for( auto& [eid, entity] : m_entities) {
                    if ( entity.hasComponents(key) ) {
                        entities.push_back(eid);
                    }
                }

                return entities;
            }

            template<typename C>
            C* add(entity_t entity_id) {
                auto& entity = m_entities.at(entity_id);
                auto comp = entity.template add<C>();

                m_types.fireAdd(*this, entity_id, Component<C>::typeID());
                return comp;
            }

            template<typename C>
            C* set(entity_t entity_id, const C* ptr) {
                auto& entity = m_entities.at(entity_id);
                auto comp = entity.set<C>(ptr);

                m_types.fireAdd(*this, entity_id, Component<C>::typeID());
                return comp;
            }

            template<typename C>
            C* get(entity_t entity_id) {
                auto& entity = m_entities.at(entity_id);
                return entity.get<C>();
            }

        private:
            TypeRegistry& m_types;
            entity_t m_next;
            std::map<entity_t, Entity> m_entities;

    };

}

#endif //FGGL_ECS3_PROTOTYPE_WORLD_H